#include "editor.h"

// Addresse des ersten angezeigten Bytes (links oben im Display)
uint16_t editor_start = 0;

// Zaehlt Position der Halbbytes
uint8_t editor_cursor_position = 0;

// Adresse nach dem letzten bearbeiteten Byte
uint16_t editor_end_pointer = 0;

// Zwischenspeicher fuer eingegebene Daten
uint8_t editor_buffer[EDITOR_MAXMEMSIZE];

// Sprungziele fuer den Cursor; Beginn der einzelnen Bytes
const uint8_t editor_tabs[] = {7, 9, 12, 14, 18, 20, 23, 25};

// Sprungziele fuer den Cursor; Beginn der einzelnen Half-Bytes
const uint8_t editor_real_pos[] = {7, 8, 9, 10, 12, 13, 14, 15,
        18, 19, 20, 21, 23, 24, 25, 26};

/* 
 * Inititalisiert den Editor mit einem leeren Puffer
 */
void editor_init(void) {
    uint16_t i = 0;

    // Jedes Zeichen im Puffer auf 0xff setzen (leeren)
    for (i = 0; i < EDITOR_MAXMEMSIZE; i++) {
        editor_buffer[i] = 0xff;
    }

}

/*
 * Setzt den Cursor des LCD an die Position, auf die 
 * editor_cursor zeigt.
 */
void editor_update_cursor(void) {

    // Berechnet die Zeile und die Position in der Zeile und 
    // setzt den Cursor dorthin.
    lcd_set_position(editor_cursor_position / 16 * 27 +
            editor_real_pos[editor_cursor_position % 16]);
}

/*
 * Zeichnet den Editorinhalt der aktuellen Stelle.
 */
void editor_display(void) {

    // Speichert LCD-Zeilen-Konstanten fuer einfachen Zugriff
    uint8_t line[] = {LCD_LINE1, LCD_LINE2, LCD_LINE3,
            LCD_LINE4};

    uint8_t j = 0;

    // Zeichne Zeilen 1-3 (in der 4. stehen die Menuepunkte)
    for (j = 0; j < 3; j++) {

        // Initialisiere eine Zeile der Editoransicht und
        // trage Adresse (links) und Daten (rechts) ein
        char str[28] = "   h:                      ";
        str[0] = misc_halfbyte_to_char(
                (editor_start + j * 8)>>8);
        str[1] = misc_halfbyte_to_char(
                (editor_start + j * 8)>>4);
        str[2] = misc_halfbyte_to_char(
                editor_start + j * 8);

        uint8_t i = 0;

        for (i = 0; i < 8; i++) {
            str[editor_tabs[i]] = misc_halfbyte_to_char(
                    editor_buffer[editor_start + j * 8 + i]>>4);
            str[editor_tabs[i] + 1] = misc_halfbyte_to_char(
                    editor_buffer[editor_start + j * 8 + i]);
        }

        // Gib den String im LCD aus.
        lcd_write_string(line[j], str);
    }

    // Setzt den Cursor
    editor_update_cursor();
}

/*
 * Wertet Tasteneingaben aus.
 */
void editor_handle_keystroke(uint8_t key) {

    switch (key) {

        // Je nach Cursorposition: Ein Halfbyte nach links 
        // oder eine Zeile nach oben und ganz nach rechts
        case KEY_LEFT:

            if (editor_cursor_position != 0) {
                // Cursor ist noch nicht links oben, also
                // Cursor eins nach vorne und anzeigen.
                editor_cursor_position--;
                editor_update_cursor();
            } else if (editor_start != 0) {
                // Cursor ist links oben und es gibt noch
                // Zeilen weiter oben. Dann: Eine Zeile nach 
                // oben scrollen und Cursor aktualisieren.
                editor_start = editor_start - 8;
                editor_cursor_position = 15;
                editor_display();
            }

            break;

        // Je nach Cursorposition: ein Halfbyte nach rechts 
        // oder herunterscrollen und Cursor an den Anfang der 
        // nun letzten Zeile setzen.
        case KEY_RIGHT:

            if (editor_cursor_position != 47) { 
                // Nicht rechts unten: Cursor eins nach hinten
                editor_cursor_position++;
                editor_update_cursor();
            } else if (EDITOR_MAXMEMSIZE - 24 > editor_start) {
                 // Cursor ist rechts unten und es gibt noch
                 // Zeilen weiter unten: Eine Zeile nach unten 
                 // scrollen, Cursor setzen und aktualisieren.
                editor_start = editor_start + 8;
                editor_cursor_position = 32;
                editor_display();
            }

            break;

        // Cursor eine Zeile nach oben, ggf. nach oben scrollen.
        case KEY_UP:

            if (editor_cursor_position / 16 > 0) {
                // Cursor ist nicht in oberster Zeile:
                // Cursor setzen und anzeigen
                editor_cursor_position = 
                    editor_cursor_position - 16;
                editor_update_cursor();
            } else if (editor_start > 0) {
                // Cursor ist in oberster Zeile und es gibt 
                // noch Zeilen weiter oben. Dann: Hochscrollen.
                editor_start = editor_start - 8;
                editor_display();
            }

            break;

        // Eine Zeile nach unten, ggf. nach unten scrollen.
        case KEY_DOWN:

            if (editor_cursor_position / 16 < 2) {
                // Cursor ist nicht in unterster Zeile: 
                // Cursor setzen und anzeigen
                editor_cursor_position = editor_cursor_position
                        + 16;
                editor_update_cursor();
            } else if (EDITOR_MAXMEMSIZE - 24 > editor_start) {
                // Cursor ist in unterster Zeile und es gibt noch
                // Zeilen weiter unten. Dann: Herunterscrollen.
                editor_start = editor_start + 8;
                editor_display();
            }

            break;

        // Zeichen in LCD und den Puffer schreiben und Cursor
        // weiterruecken.
        case KEY_0:
        case KEY_1:
        case KEY_2:
        case KEY_3:
        case KEY_4:
        case KEY_5:
        case KEY_6:
        case KEY_7:
        case KEY_8:
        case KEY_9:
        case KEY_A:
        case KEY_B:
        case KEY_C:
        case KEY_D:
        case KEY_E:
        case KEY_F:

            { 
                // Aktuelle Position im Puffer
                uint16_t pos = editor_start +
                        editor_cursor_position / 2;

                if (editor_cursor_position % 2 == 0) {
                    // Linkes Half-Byte speichern
                    editor_buffer[pos] = (key<<4) |
                            (editor_buffer[pos] & 0x0f);
                } else { 
                    // Rechtes Half-Byte speichern
                    editor_buffer[pos] = key |
                            (editor_buffer[pos] & 0xf0);
                }

                // Endpointer neu setzen, wenn ein anderes Zeichen
                // als 0xff hinter dem aktuellen Endpointer
                // geschrieben wurde.
                if (key != KEY_F && pos + 1 >
                        editor_end_pointer) {
                    editor_end_pointer = pos + 1;
                }

                // Zeichen auf LCD schreiben
                lcd_write_char(editor_cursor_position / 16 * 27
                        + editor_real_pos[editor_cursor_position
                        % 16], misc_halfbyte_to_char(key));

                // Eine Position weiterruecken
                editor_handle_keystroke(KEY_RIGHT);
                break;
            }

        // Eine Editorseite (drei Zeilen) nach unten scrollen.
        case KEY_PDOWN:

            // Je nach Anzahl der noch vorhandenen Zeilen im 
            // Puffer versuchen, drei oder die max. noch 
            // vorhandene Zeilenzahl herunterzuscrollen
            if (EDITOR_MAXMEMSIZE - 40 > editor_start) {
                editor_start = editor_start + 24;
                editor_display();
            } else if (EDITOR_MAXMEMSIZE - 32 > editor_start) {
                editor_start = editor_start + 16;
                editor_display();
            } else if (EDITOR_MAXMEMSIZE - 24 > editor_start) {
                editor_start = editor_start + 8;
                editor_display();
            }

            break;

        // Eine Editorseite (drei Zeilen) nach unten scrollen.
        case KEY_PUP:

            // Je nach Anzahl der noch vorhandenen Zeilen im 
            // Puffer versuchen, drei oder die max. noch 
            // vorhandene Zeilenzahl hinaufzuscrollen
            if (editor_start > 16) {
                editor_start = editor_start - 24;
                editor_display();
            } else if (editor_start > 8) {
                editor_start = editor_start - 16;
                editor_display();
            } else if (editor_start > 0) {
                editor_start = editor_start - 8;
                editor_display();
            }

            break;

        // Unbekannte Taste, nichts tun.
        default:
            break;

    }

}

/*
 * Editor leeren und alle Variablen zuruecksetzen.
 */
void editor_clear(void) {
    uint16_t i = 0;

    // Puffer bis zum End Pointer (nach dem nur noch 0xFF
    // folgt) wieder mit 0xFF fuellen
    for (i = 0; i < editor_end_pointer; i++) {
        editor_buffer[i] = 0xff;
    }

    // Alles zuruecksetzen.
    editor_end_pointer = 0;
    editor_start = 0;
    editor_cursor_position = 0;
}

/*
 * EPROM in den Puffer lesen.
 */
void editor_read_eprom(void) {
    editor_end_pointer = 0;
    uint16_t i = 0;

    // EPROM auslesen und Endpointer neu setzen
    for (i = 0; i < EDITOR_MAXMEMSIZE; i++) {
        editor_buffer[i] = eprom_read(i);

        if (editor_buffer[i] != 0xff) {
            editor_end_pointer = i + 1;
        }

    }

    // Anzeige auf Anfang setzen.
    editor_start = 0;
    editor_cursor_position = 0;
}

/* 
 * Brennt den Inhalt des Puffers in das EPROM.
 */
void editor_burn(void) {

    // Brennvorgang initialisieren.
    eprom_write_start();
    uint16_t i = 0;

    // EPROM brennen, dabei 0xff auslassen.
    for (i = 0; i < editor_end_pointer; i++) {

        if (editor_buffer[i] != 0xff) {
            eprom_write(i, editor_buffer[i]);
        }

    }

    // Brennvorgang beenden.
    eprom_write_stop();
}
