#include "lcd.h"

/*
    Speichert den aktuellen Status des Cursors. Ist Bit 0
    gesetzt, so wird der Cursor angezeigt, ansonsten nicht. Ist
    Bit 1 gesetzt, so wird der Cursor im unteren Controller
    angezeigt, sonst im oberen.
*/
uint8_t lcd_cursor = 0;

/*
 * Liest den Wert des uebergebenen Registers des uebergebenen
 * Controllers aus und gibt ihn zurueck.
 */
uint8_t lcd_read(uint8_t controller, uint8_t reg) {
    uint8_t value = 0;

    // Das LCD soll auf den LCD-/EPROM-Bus schreiben.
    misc_set_data_direction(MISC_DD_LCD);

    switch (reg) {

        // Aus dem Befehlsregister soll gelesen werden.
        case LCD_IR:
            PORTC &= ~(1<<RS);
            break;

        // Aus dem Datenregister soll gelesen werden.
        case LCD_DR:
            PORTC |= (1<<RS);
            break;

        // Ein falscher Wert wurde uebergeben.
        default:
            misc_error();
            break;

    }

    switch (controller) {

        // Aus dem oberen Controller soll gelesen werden.
        case LCD_CTL_TOP:

            // Oberen Controller aktivieren, Stabilisierung
            // abwarten, den gelesenen Wert speichern und den
            // Controller wieder deaktivieren.
            PORTC |= (1<<E1);
            _delay_us(10);
            value = PINA;
            PORTC &= ~(1<<E1);
            break;

        // Aus dem unteren Controller soll gelesen werden.
        case LCD_CTL_BOTTOM:

            // Unteren Controller aktivieren, Stabilisierung
            // abwarten, den gelesenen Wert speichern und den
            // Controller wieder deaktivieren.
            PORTC |= (1<<E2);
            _delay_us(10);
            value = PINA;
            PORTC &= ~(1<<E2);
            break;

        // Ein falscher Wert wurde uebergeben.
        default:
            misc_error();
            break;

    }

    return value;
}

/*
 * Blockt, bis das Busy-Flag des uebergebenen Controllers nicht
 * mehr gesetzt ist.
 */
void lcd_check_busy(uint8_t controller) {

    while (lcd_read(controller, LCD_IR) & LCD_BUSY);

}

/*
 * Schreibt value in das uebergebene Register des 
 * uebergebenen Controllers.
 */
void lcd_write(uint8_t controller, uint8_t reg, uint8_t value) {

    // Der Controller soll auf den LCD-/EPROM-Bus schreiben.
    misc_set_data_direction(MISC_DD_UC);

    switch (reg) {

        // Schreibziel: Befehlsregister
        case LCD_IR:
            PORTC &= ~(1<<RS);
            break;

        // Schreibziel: Datenregister
        case LCD_DR:
            PORTC |= (1<<RS);
            break;

        // Ein falscher Wert wurde uebergeben.
        default:
            misc_error();
            break;

    }

    // Den Wert auf die Datenleitung legen.
    PORTA = value;

    switch (controller) {

        // In den oberen Controller soll geschrieben werden.
        case LCD_CTL_TOP:

            //  Oberen Controller aktivieren, Stabilisierung
            // abwarten und den Controller wieder deaktivieren.
            PORTC |= (1<<E1);
            _delay_us(10);
            PORTC &= ~(1<<E1);
            break;

        // In den unteren Controller soll geschrieben werden.
        case LCD_CTL_BOTTOM:

            // Unteren Controller aktivieren, Stabilisierung
            // abwarten und den Controller wieder deaktivieren.
            PORTC |= (1<<E2);
            _delay_us(10);
            PORTC &= ~(1<<E2);
            break;

        // Ein falscher Wert wurde uebergeben.
        default:
            misc_error();
            break;

    }

    // Warten, bis der Controller fertig ist.
    lcd_check_busy(controller);

}

/*
 * Initialisiert das LCD mit keinem Text und keinem Cursor.
 */
void lcd_init() {

    // Warten, bis das LCD nach dem Einschalten bereit ist.
    _delay_ms(20);

    // Beide Controller auf 8-Bit-Datenbreite setzen, dabei
    // zuerst auf ein-zeilig, dann auf zwei-zeilig setzen.
    lcd_write(LCD_CTL_TOP, LCD_IR, 0x30);
    lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x30);
    lcd_write(LCD_CTL_TOP, LCD_IR, 0x38);
    lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x38);

    // Die Anzeige leeren.
    lcd_write(LCD_CTL_TOP, LCD_IR, 0x01);
    lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x01);

    // Den Cursor an den Anfang setzen.
    lcd_write(LCD_CTL_TOP, LCD_IR, 0x02);
    lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x02);

    // Beim Schreiben eines Zeichens den Cursor nach rechts
    // verschieben statt der ganzen Anzeige.
    lcd_write(LCD_CTL_TOP, LCD_IR, 0x06);
    lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x06);

    // Den Cursor abschalten.
    lcd_write(LCD_CTL_TOP, LCD_IR, 0x0c);
    lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x0c);
}

/*
 * Setzt die Adresse (und damit die Position des Cursors) auf
 * die uebergebene Position, um anschliessend dort schreiben zu
 * koennen. Die Position muss zwischen 0 und 107 (4 * 27 - 1)
 * liegen und wird umgerechnet auf die richtige Position im
 * oberen bzw. unteren Controller.
 */
void lcd_set_position(uint8_t pos) {
    uint8_t ctl = 0;

    if (pos < 27) { // Position ist in Zeile 1

        // Der richtige Controller ist der obere, Umrechnung
        // der Position nicht noetig.
        ctl = LCD_CTL_TOP;

    } else if (pos < 54) { // Position ist in Zeile 2

        // Der richtige Controller ist der obere, Position muss
        // umgerechnet werden (pos - 27 entspricht der Position
        // in der Zeile, 64 ist der Offset fuer die 2. Zeile).
        ctl = LCD_CTL_TOP;
        pos = pos - 27 + 64;

    } else if (pos < 81) { // Position ist in Zeile 3

        // Der richtige Controller ist der untere, Position
        // muss umgerechnet werden (pos - 54 entspricht der
        // Position in der Zeile, kein Offset fuer die 1.
        // Zeile).
        ctl = LCD_CTL_BOTTOM;
        pos = pos - 54;

    } else if (pos < 108) { // Position ist in Zeile 4

        // Der richtige Controller ist der untere, Position
        // muss umgerechnet werden (pos - 81 entspricht der
        // Position in der Zeile, 64 ist der Offset fuer die 2.
        // Zeile).
        ctl = LCD_CTL_BOTTOM;
        pos = pos - 81 + 64;

    } else { // keine gueltige Position
        misc_error();
    }

    // Schreibe die umgerechnete Position in den richtigen
    // Controller (Bit 7 gehoert nicht zur eigentlichen
    // Position, sondern ist der Befehl zum Schreiben der
    // Position und muss 1 sein).
    lcd_write(ctl, LCD_IR, pos | 0x80);

    // Falls der Cursor angezeigt wird, muss geprueft werden,
    // ob der Cursor aufgrund der neuen Position den Controller
    // wechseln muss.
    if (lcd_cursor & 0x01) {

        if (lcd_cursor & 0x02 && ctl == LCD_CTL_TOP) {
            
            // Der Cursor wird im unteren Controller angezeigt,
            // die neue Position befindet sich aber im oberen.
            // Dann: Speichere den Status, deaktiviere den 
            // Cursor im  unteren und aktiviere ihn im oberen 
            // Controller.
            lcd_cursor = 0x01;
            lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x0c);
            lcd_write(LCD_CTL_TOP, LCD_IR, 0x0e);

        } else if (!(lcd_cursor & 0x02) &&
                ctl == LCD_CTL_BOTTOM) {
            // Der Cursor wird im oberen Controller angezeigt,
            // die neue Position befindet sich aber im unteren.
            // Dann: Speichere den Status, deaktiviere den 
            // Cursor im oberen und aktiviere ihn im unteren 
            // Controller.
            lcd_cursor = 0x03;
            lcd_write(LCD_CTL_TOP, LCD_IR, 0x0c);
            lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x0e);

        }

    }

}

/*
 * Schreibt das uebergebene Zeichen an die angegebene Position.
 */
void lcd_write_char(uint8_t pos, uint8_t ch) {
    lcd_set_position(pos);

    // Waehle Controller entsprechend Cursorposition
    if (pos < 54) {
        lcd_write(LCD_CTL_TOP, LCD_DR, ch);
    } else {
        lcd_write(LCD_CTL_BOTTOM, LCD_DR, ch);
    }

}

/*
 * Schreibt den uebergebenen String an die angegebene Position.
 */
void lcd_write_string(uint8_t pos, char str[]) {
    uint8_t i = 0;

    // Schreibe das i-te Zeichen an die (pos + i)-te Position.
    // 0x00 markiert das Ende einer Zeichenkette.
    while (str[i] != 0) {
        lcd_write_char(pos + i, str[i]);
        i++;
    }

}

/*
 * Leert das LCD.
 */
void lcd_clear(void) {
    lcd_write(LCD_CTL_TOP, LCD_IR, 0x01);
    lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x01);
}

// Aktiviert oder deaktiviert den Cursor.
void lcd_display_cursor(uint8_t enable) {

    if (enable) { 
        // Auswahl des Controllers, in dem der Cursor aktiviert
        // werden soll:
        if (lcd_cursor & 0x02) {
            
            // unterer Controller
            lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x0e);
            lcd_cursor = 0x03;

        } else {
            
            // oberer Controller
            lcd_write(LCD_CTL_TOP, LCD_IR, 0x0e);
            lcd_cursor = 0x01;
        
        }

    } else { 
        // Auswahl des Controllers, in dem der Cursor 
        // deaktiviert werden soll:

        if (lcd_cursor & 0x02) {
            
            // unterer Controller
            lcd_write(LCD_CTL_BOTTOM, LCD_IR, 0x0c);
            lcd_cursor = 0x02;
        } else {
            
            // oberer Controller
            lcd_write(LCD_CTL_TOP, LCD_IR, 0x0c);
            lcd_cursor = 0x00;
        }

    }

}
