#include "tastencontroller.h"

// Maske fuer die Tastenkabel, die an Port A liegen.
const uint8_t WIRES_AT_A = (1<<WIRE_ORANGE) | (1<<WIRE_PURPLE);

// Maske fuer die Tastenkabel, die an Port B liegen.
const uint8_t WIRES_AT_B = (1<<WIRE_GREEN) | (1<<WIRE_RED)
        | (1<<WIRE_BLUE) | (1<<WIRE_BROWN) | (1<<WIRE_YELLOW)
        | (1<<WIRE_WHITE);

// Maske fuer die Tastengruppen, die an Port D liegen.
const uint8_t WIRES_AT_D = (1<<WIRE_GROUP1) | (1<<WIRE_GROUP2)
        | (1<<WIRE_GROUP3) | (1<<WIRE_GROUP4);

/*
    Array der Tastengruppen, um besser ueber diese iterieren zu
    koennen.
*/
const uint8_t groups[] = {WIRE_GROUP1, WIRE_GROUP2, WIRE_GROUP3,
        WIRE_GROUP4};

/*
    Array der Tasten, um besser ablesen zu koennen, welche
    gedrueckt wurde.
*/
const uint8_t keys[] = {KEY_0, KEY_1, KEY_2, KEY_3, KEY_4,
        KEY_5, KEY_6, KEY_7, KEY_8, KEY_9, KEY_A, KEY_B, KEY_C,
        KEY_D, KEY_E, KEY_F, KEY_LEFT, KEY_RIGHT, KEY_DOWN,
        KEY_UP, KEY_SK1, KEY_SK2, KEY_SK3, KEY_SK4, KEY_OK,
        KEY_PDOWN, KEY_PUP, KEY_NONE, KEY_NONE, KEY_NONE,
        KEY_NONE, KEY_NONE};

// Diese Taste wurde gedrueckt.
uint8_t pressed_key = KEY_NONE;

// Diese Taste wurde auch wieder losgelassen.
uint8_t typed_key = KEY_NONE;

/*
    Prueft die uebergebene Gruppe auf eine gedrueckte Taste und
    gibt die Nummer des Kabels zurueck.
*/
uint8_t check_group(uint8_t wire_group) {
    // 0 an die gegebene Gruppe legen
    DDRD |= (1<<wire_group);
    // Stabilisierung abwarten
    _delay_us(10);
    // Speichern, an welchen Tastenkabeln die 0 ankommt
    uint8_t b = PINB | ~(WIRES_AT_B);
    uint8_t a = PINA | ~(WIRES_AT_A);
    // 0 wieder wegnehmen
    DDRD &= ~(WIRES_AT_D);

    if (b < 0xff || a < 0xff) {
            // An einem der Tastenkabeln kam die 0 an.

        if ((b & (1<<WIRE_GREEN)) == 0x00) { // Am gruenen
            return 1;
        } else if ((b & (1<<WIRE_RED)) == 0x00) { // Am roten
            return 2;
        } else if ((b & (1<<WIRE_BLUE)) == 0x00) { // Am blauen
            return 3;
        } else if ((b & (1<<WIRE_BROWN)) == 0x00) {
                // Am braunen
            return 4;
        } else if ((b & (1<<WIRE_YELLOW)) == 0x00) {
                // Am gelben
            return 5;
        } else if ((b & (1<<WIRE_WHITE)) == 0x00) {
                // Am weissen
            return 6;
        } else if ((a & (1<<WIRE_PURPLE)) == 0x00) {
                // Am lilanen
            return 7;
        } else if ((a & (1<<WIRE_ORANGE)) == 0x00) {
                // Am orangenen
            return 8;
        }

    }

    // Keine Taste wurde gedrueckt
    return 0;
}

/*
    Prueft alle Tastengruppen auf eine gedrueckte Taste und gibt
    diese zurueck.
*/
uint8_t check_keystroke(void) {
    // Nummer des Kabels der gedrueckten Taste
    uint8_t wire;
    // Nummer der aktuellen Tastengruppe
    uint8_t group;

    /*
        Alle Gruppen auf eine gedrueckte Taste ueberpruefen,
        vorzeitig abbrechen, sobald eine gefunden wurde
    */
    for (group = 0; group < 4; group++) {
        wire = check_group(groups[group]);

        if (wire != 0) { // Eine Taste wurde gedrueckt
            break;
        }

    }

    if (wire != 0) { // Eine Taste wurde gedrueckt

        /*
            Die zu der Gruppe und dem Kabel gehoerende Taste
            zurueckgeben
        */
        return keys[group * 8 + wire - 1];
    } else { // Keine Taste wurde gedrueckt
        return KEY_NONE;
    }

}

/*
    Initialisiert den Tastencontroller. Setzt Datenrichtungen
    und Pull-Up-Widerstaende der Ports und initialisiert USI.
*/
void init(void) {
    // Initialisiert USI
    usi_init();

    /*
        Setzt die Datenrichtungen fuer die Tastenkabel und
        -gruppen, setzt sie auf lesend. Beim Ueberpruefen auf
        Tastendruecke wird Port D teilweise kurzzeitig auf
        schreibend gesetzt.
    */
    DDRA &= ~(WIRES_AT_A);
    DDRB &= ~(WIRES_AT_B);
    DDRD &= ~(WIRES_AT_D);

    /*
        Aktiviert die Pull-Up-Widerstaende an den Tastenkabeln
        und deaktiviert sie fuer die Tastengruppen.
    */
    PORTA |= WIRES_AT_A;
    PORTB |= WIRES_AT_B;
    PORTD &= ~(WIRES_AT_D);
}

/*
    Wird beim Starten des Controllers aufgerufen. Initialisiert
    den Controller und bearbeitet anschliessend in einer Endlos-
    Schleife Tastendruecke und Anfragen des Hauptcontrollers.
*/
int main(void) {
    // Initialisierung
    init();

    uint8_t usi_process_return;
    uint8_t i = 0;
    uint8_t j = 0;

    // Endlosschleife
    while (1) {

        if (pressed_key == KEY_NONE) {
                // Noch keine Taste gedrueckt

            // Auf Tastendruck ueberpruefen
            pressed_key = check_keystroke();
            // Hilfsvariable zuruecksetzen
            i = 0;
        } else if (typed_key == KEY_NONE) {
                // Taste wurde noch nicht losgelassen

            /*
                ~20 ms warten (Entprellung), dazwischen USI
                bearbeiten (Fehlendes Delay passiert in
                usi_process()
            */
            if (i < 2) {
                _delay_ms(1);
                i++;
            } else if (check_keystroke() == KEY_NONE) {
                    // Taste wurde losgelassen
                typed_key = pressed_key;
            }

        }

        // USI bearbeiten und Rueckgabe speichern
        usi_process_return = usi_process();

        if (usi_process_return == USI_NOP) {
                // Nichts interessantes passiert

            // Timeout-Zaehler erhoehen
            j++;

            if (j > USI_TIMEOUT) {
                    // Timeout erreicht, Fehler annehmen

                // USI zuruecksetzen
                j = 0;
                usi_reset();
            }

        } else { // Etwas interessantes passiert
            // Timeout zuruecksetzen
            j = 0;

            if (usi_process_return == USI_SET_DATA) {
                    // Taste muss an USI gegeben werden

                // Gedrueckte Taste uebergeben
                usi_set_data(typed_key);

                if (typed_key != KEY_NONE) {
                        // Eine Taste war getippt worden

                    // Tastendruecke zuruecksetzen
                    pressed_key = KEY_NONE;
                    typed_key = KEY_NONE;
                }

            } else if (usi_process_return == USI_GET_DATA) {
                    // Ein Wert wurde empfangen

                /*
                    Fehler annehmen, da der Tastencontroller nur
                    sendet.
                */
                usi_reset();
            }

        }

    }

    return 0;
}
